@@ -12,17 +12,22 @@ files on disk), and we parse the dependency structure as a tree rather than
  a flat list.
 """
 
-import os
-import glob
-import errno
-import logging
-import functools
-from typing import List, Dict, Optional, Any
+from __future__ import annotations
 
-from elftools.elf.elffile import ELFFile  # type: ignore
+import errno
+import functools
+import glob
+import logging
+import os
+from pathlib import Path
+from typing import Any
+
+from elftools.elf.elffile import ELFFile
+
+from .libc import Libc, get_libc
 
 log = logging.getLogger(__name__)
-__all__ = ['lddtree']
+__all__ = ["lddtree"]
 
 
 def normpath(path: str) -> str:
@@ -33,7 +38,7 @@ def normpath(path: str) -> str:
     //..// -> //
     //..//..// -> ///
     """
-    return os.path.normpath(path).replace('//', '/')
+    return os.path.normpath(path).replace("//", "/")
 
 
 def readlink(path: str, root: str, prefixed: bool = False) -> str:
@@ -59,9 +64,9 @@ def readlink(path: str, root: str, prefixed: bool = False) -> str:
     -------
     A fully resolved symlink path
     """
-    root = root.rstrip('/')
+    root = root.rstrip("/")
     if prefixed:
-        path = path[len(root):]
+        path = path[len(root) :]
 
     while os.path.islink(root + path):
         path = os.path.join(os.path.dirname(path), os.readlink(root + path))
@@ -69,13 +74,13 @@ def readlink(path: str, root: str, prefixed: bool = False) -> str:
     return normpath((root + path) if prefixed else path)
 
 
-def dedupe(items: List[str]) -> List[str]:
+def dedupe(items: list[str]) -> list[str]:
     """Remove all duplicates from ``items`` (keeping order)"""
-    seen = {}  # type: Dict[str, str]
+    seen: dict[str, str] = {}
     return [seen.setdefault(x, x) for x in items if x not in seen]
 
 
-def parse_ld_paths(str_ldpaths, root='', path=None) -> List[str]:
+def parse_ld_paths(str_ldpaths: str, path: str, root: str = "") -> list[str]:
     """Parse the colon-delimited list of paths and apply ldso rules to each
 
     Note the special handling as dictated by the ldso:
@@ -96,24 +101,21 @@ def parse_ld_paths(str_ldpaths, root='', path=None) -> List[str]:
     -------
         list of processed paths
     """
-    ldpaths = []  # type: List[str]
-    for ldpath in str_ldpaths.split(':'):
-        if ldpath == '':
+    ldpaths: list[str] = []
+    for ldpath in str_ldpaths.split(":"):
+        if ldpath == "":
             # The ldso treats "" paths as $PWD.
             ldpath = os.getcwd()
-        elif '$ORIGIN' in ldpath:
-            ldpath = ldpath.replace('$ORIGIN',
-                                    os.path.dirname(os.path.abspath(path)))
+        elif "$ORIGIN" in ldpath:
+            ldpath = ldpath.replace("$ORIGIN", os.path.dirname(os.path.abspath(path)))
         else:
             ldpath = root + ldpath
         ldpaths.append(normpath(ldpath))
     return [p for p in dedupe(ldpaths) if os.path.isdir(p)]
 
 
-@functools.lru_cache()
-def parse_ld_so_conf(ldso_conf: str,
-                     root: str = '/',
-                     _first: bool = True) -> List[str]:
+@functools.lru_cache
+def parse_ld_so_conf(ldso_conf: str, root: str = "/", _first: bool = True) -> list[str]:
     """Load all the paths from a given ldso config file
 
     This should handle comments, whitespace, and "include" statements.
@@ -131,27 +133,25 @@ def parse_ld_so_conf(ldso_conf: str,
     -------
     list of paths found
     """
-    paths = []  # type: List[str]
+    paths: list[str] = []
 
-    dbg_pfx = '' if _first else '  '
+    dbg_pfx = "" if _first else "  "
     try:
-        log.debug('%sparse_ld_so_conf(%s)', dbg_pfx, ldso_conf)
+        log.debug("%sparse_ld_so_conf(%s)", dbg_pfx, ldso_conf)
         with open(ldso_conf) as f:
             for line in f.readlines():
-                line = line.split('#', 1)[0].strip()
+                line = line.split("#", 1)[0].strip()
                 if not line:
                     continue
-                if line.startswith('include '):
+                if line.startswith("include "):
                     line = line[8:]
-                    if line[0] == '/':
-                        line = root + line.lstrip('/')
+                    if line[0] == "/":
+                        line = root + line.lstrip("/")
                     else:
-                        line = os.path.dirname(ldso_conf) + '/' + line
-                    log.debug('%s  glob: %s', dbg_pfx, line)
+                        line = os.path.dirname(ldso_conf) + "/" + line
+                    log.debug("%s  glob: %s", dbg_pfx, line)
                     for path in glob.glob(line):
-                        paths += parse_ld_so_conf(path,
-                                                  root=root,
-                                                  _first=False)
+                        paths += parse_ld_so_conf(path, root=root, _first=False)
                 else:
                     paths += [normpath(root + line)]
     except OSError as e:
@@ -166,8 +166,8 @@ def parse_ld_so_conf(ldso_conf: str,
     return paths
 
 
-@functools.lru_cache()
-def load_ld_paths(root: str = '/', prefix: str = '') -> Dict[str, List[str]]:
+@functools.lru_cache
+def load_ld_paths(root: str = "/", prefix: str = "") -> dict[str, list[str]]:
     """Load linker paths from common locations
 
     This parses the ld.so.conf and LD_LIBRARY_PATH env var.
@@ -183,28 +183,48 @@ def load_ld_paths(root: str = '/', prefix: str = '') -> Dict[str, List[str]]:
     -------
     dict containing library paths to search
     """
-    ldpaths = {'conf': [], 'env': [], 'interp': []}  # type: Dict
+    ldpaths: dict = {"conf": [], "env": [], "interp": []}
 
     # Load up $LD_LIBRARY_PATH.
-    env_ldpath = os.environ.get('LD_LIBRARY_PATH')
+    env_ldpath = os.environ.get("LD_LIBRARY_PATH")
     if env_ldpath is not None:
-        if root != '/':
-            log.warning('ignoring LD_LIBRARY_PATH due to ROOT usage')
+        if root != "/":
+            log.warning("ignoring LD_LIBRARY_PATH due to ROOT usage")
         else:
             # XXX: If this contains $ORIGIN, we probably have to parse this
             # on a per-ELF basis so it can get turned into the right thing.
-            ldpaths['env'] = parse_ld_paths(env_ldpath, path='')
+            ldpaths["env"] = parse_ld_paths(env_ldpath, path="")
 
-    # Load up /etc/ld.so.conf.
-    ldpaths['conf'] = parse_ld_so_conf(root + prefix + '/etc/ld.so.conf',
-                                       root=root)
-    # the trusted directories are not necessarily in ld.so.conf
-    ldpaths['conf'].extend(['/lib', '/lib64/', '/usr/lib', '/usr/lib64'])
-    log.debug('linker ldpaths: %s', ldpaths)
+    libc = get_libc()
+    if libc == Libc.MUSL:
+        # from https://git.musl-libc.org/cgit/musl/tree/ldso
+        # /dynlink.c?id=3f701faace7addc75d16dea8a6cd769fa5b3f260#n1063
+        root_prefix = Path(root) / prefix
+        ld_musl = list((root_prefix / "etc").glob("ld-musl-*.path"))
+        assert len(ld_musl) <= 1
+        if len(ld_musl) == 0:
+            ldpaths["conf"] = [
+                root + "/lib",
+                root + "/usr/local/lib",
+                root + "/usr/lib",
+            ]
+        else:
+            ldpaths["conf"] = []
+            for ldpath in ld_musl[0].read_text().split(":"):
+                ldpath_stripped = ldpath.strip()
+                if ldpath_stripped == "":
+                    continue
+                ldpaths["conf"].append(root + ldpath_stripped)
+    else:
+        # Load up /etc/ld.so.conf.
+        ldpaths["conf"] = parse_ld_so_conf(root + prefix + "/etc/ld.so.conf", root=root)
+        # the trusted directories are not necessarily in ld.so.conf
+        ldpaths["conf"].extend(["/lib", "/lib64/", "/usr/lib", "/usr/lib64"])
+    log.debug("linker ldpaths: %s", ldpaths)
     return ldpaths
 
 
-def compatible_elfs(elf1, elf2):
+def compatible_elfs(elf1: ELFFile, elf2: ELFFile) -> bool:
     """See if two ELFs are compatible
 
     This compares the aspects of the ELF to see if they're compatible:
@@ -219,20 +239,29 @@ def compatible_elfs(elf1, elf2):
     -------
     True if compatible, False otherwise
     """
-    osabis = frozenset([e.header['e_ident']['EI_OSABI'] for e in (elf1, elf2)])
-    compat_sets = (frozenset('ELFOSABI_%s' % x
-                             for x in ('NONE',
-                                       'SYSV',
-                                       'GNU',
-                                       'LINUX', )), )
-    return ((len(osabis) == 1 or
-             any(osabis.issubset(x)
-                 for x in compat_sets)) and elf1.elfclass == elf2.elfclass and
-            elf1.little_endian == elf2.little_endian and
-            elf1.header['e_machine'] == elf2.header['e_machine'])
+    osabis = frozenset(e.header["e_ident"]["EI_OSABI"] for e in (elf1, elf2))
+    compat_sets = (
+        frozenset(
+            "ELFOSABI_%s" % x
+            for x in (
+                "NONE",
+                "SYSV",
+                "GNU",
+                "LINUX",
+            )
+        ),
+    )
+    return (
+        (len(osabis) == 1 or any(osabis.issubset(x) for x in compat_sets))
+        and elf1.elfclass == elf2.elfclass
+        and elf1.little_endian == elf2.little_endian
+        and elf1.header["e_machine"] == elf2.header["e_machine"]
+    )
 
 
-def find_lib(elf, lib, ldpaths, root='/'):
+def find_lib(
+    elf: ELFFile, lib: str, ldpaths: list[str], root: str = "/"
+) -> tuple[str | None, str | None]:
     """Try to locate a ``lib`` that is compatible to ``elf`` in the given
     ``ldpaths``
 
@@ -242,7 +271,7 @@ def find_lib(elf, lib, ldpaths, root='/'):
         The elf which the library should be compatible with (ELF wise)
     lib : str
         The library (basename) to search for
-    ldpaths : List[str]
+    ldpaths : list[str]
         A list of paths to search
     root : str
        The root path to resolve symlinks
@@ -257,7 +286,7 @@ def find_lib(elf, lib, ldpaths, root='/'):
         target = readlink(path, root, prefixed=True)
 
         if os.path.exists(target):
-            with open(target, 'rb') as f:
+            with open(target, "rb") as f:
                 libelf = ELFFile(f)
                 if compatible_elfs(elf, libelf):
                     return (target, path)
@@ -265,13 +294,16 @@ def find_lib(elf, lib, ldpaths, root='/'):
     return (None, None)
 
 
-def lddtree(path: str,
-            root: str = '/',
-            prefix: str = '',
-            ldpaths: Optional[Dict[str, List[str]]] = None,
-            display: Optional[str] = None,
-            _first: bool = True,
-            _all_libs: Dict = {}) -> Dict:
+def lddtree(
+    path: str,
+    root: str = "/",
+    prefix: str = "",
+    ldpaths: dict[str, list[str]] | None = None,
+    display: str | None = None,
+    exclude: frozenset[str] = frozenset(),
+    _first: bool = True,
+    _all_libs: dict = {},
+) -> dict:
     """Parse the ELF dependency tree of the specified file
 
     Parameters
@@ -289,6 +321,8 @@ def lddtree(path: str,
         will be called.
     display
         The path to show rather than ``path``
+    exclude
+        List of soname (DT_NEEDED) to exclude from the tree
     _first
         Recursive use only; is this the first ELF?
     _all_libs
@@ -318,67 +352,63 @@ def lddtree(path: str,
     if _first:
         _all_libs = {}
 
-    ret = {
-        'interp': None,
-        'path': path if display is None else display,
-        'realpath': path,
-        'needed': [],
-        'rpath': [],
-        'runpath': [],
-        'libs': _all_libs,
-    }  # type: Dict[str, Any]
+    ret: dict[str, Any] = {
+        "interp": None,
+        "path": path if display is None else display,
+        "realpath": path,
+        "needed": [],
+        "rpath": [],
+        "runpath": [],
+        "libs": _all_libs,
+    }
 
-    log.debug('lddtree(%s)' % path)
+    log.debug("lddtree(%s)" % path)
 
-    with open(path, 'rb') as f:
+    with open(path, "rb") as f:
         elf = ELFFile(f)
 
         # If this is the first ELF, extract the interpreter.
         if _first:
             for segment in elf.iter_segments():
-                if segment.header.p_type != 'PT_INTERP':
+                if segment.header.p_type != "PT_INTERP":
                     continue
 
                 interp = segment.get_interp_name()
-                log.debug('  interp           = %s', interp)
-                ret['interp'] = normpath(root + interp)
-                ret['libs'][os.path.basename(interp)] = {
-                    'path': ret['interp'],
-                    'realpath': readlink(ret['interp'],
-                                         root,
-                                         prefixed=True),
-                    'needed': [],
+                log.debug("  interp           = %s", interp)
+                ret["interp"] = normpath(root + interp)
+                ret["libs"][os.path.basename(interp)] = {
+                    "path": ret["interp"],
+                    "realpath": readlink(ret["interp"], root, prefixed=True),
+                    "needed": [],
                 }
                 # XXX: Should read it and scan for /lib paths.
-                ldpaths['interp'] = [
+                ldpaths["interp"] = [
                     normpath(root + os.path.dirname(interp)),
-                    normpath(root + prefix + '/usr' + os.path.dirname(
-                        interp).lstrip(prefix)),
+                    normpath(
+                        root + prefix + "/usr" + os.path.dirname(interp).lstrip(prefix)
+                    ),
                 ]
-                log.debug('  ldpaths[interp]  = %s', ldpaths['interp'])
+                log.debug("  ldpaths[interp]  = %s", ldpaths["interp"])
                 break
 
         # Parse the ELF's dynamic tags.
-        libs = []  # type: List[str]
-        rpaths = []  # type: List[str]
-        runpaths = []  # type: List[str]
+        libs: list[str] = []
+        rpaths: list[str] = []
+        runpaths: list[str] = []
         for segment in elf.iter_segments():
-            if segment.header.p_type != 'PT_DYNAMIC':
+            if segment.header.p_type != "PT_DYNAMIC":
                 continue
 
             for t in segment.iter_tags():
-                if t.entry.d_tag == 'DT_RPATH':
-                    rpaths = parse_ld_paths(
-                        t.rpath,
-                        root=root,
-                        path=path)
-                elif t.entry.d_tag == 'DT_RUNPATH':
-                    runpaths = parse_ld_paths(
-                        t.runpath,
-                        root=root,
-                        path=path)
-                elif t.entry.d_tag == 'DT_NEEDED':
-                    libs.append(t.needed)
+                if t.entry.d_tag == "DT_RPATH":
+                    rpaths = parse_ld_paths(t.rpath, path=path, root=root)
+                elif t.entry.d_tag == "DT_RUNPATH":
+                    runpaths = parse_ld_paths(t.runpath, path=path, root=root)
+                elif t.entry.d_tag == "DT_NEEDED":
+                    if t.needed in exclude:
+                        log.info(f"Excluding {t.needed}")
+                    else:
+                        libs.append(t.needed)
             if runpaths:
                 # If both RPATH and RUNPATH are set, only the latter is used.
                 rpaths = []
@@ -389,38 +419,47 @@ def lddtree(path: str,
         if _first:
             # Propagate the rpaths used by the main ELF since those will be
             # used at runtime to locate things.
-            ldpaths['rpath'] = rpaths
-            ldpaths['runpath'] = runpaths
-            log.debug('  ldpaths[rpath]   = %s', rpaths)
-            log.debug('  ldpaths[runpath] = %s', runpaths)
-        ret['rpath'] = rpaths
-        ret['runpath'] = runpaths
-        ret['needed'] = libs
+            ldpaths["rpath"] = rpaths
+            ldpaths["runpath"] = runpaths
+            log.debug("  ldpaths[rpath]   = %s", rpaths)
+            log.debug("  ldpaths[runpath] = %s", runpaths)
+        ret["rpath"] = rpaths
+        ret["runpath"] = runpaths
+        ret["needed"] = libs
 
         # Search for the libs this ELF uses.
-        all_ldpaths = None  # type: Optional[List[str]]
+        all_ldpaths: list[str] | None = None
         for lib in libs:
             if lib in _all_libs:
                 continue
             if all_ldpaths is None:
-                all_ldpaths = (ldpaths['rpath'] + rpaths + runpaths +
-                               ldpaths['env'] + ldpaths['runpath'] +
-                               ldpaths['conf'] + ldpaths['interp'])
+                all_ldpaths = (
+                    ldpaths["rpath"]
+                    + rpaths
+                    + runpaths
+                    + ldpaths["env"]
+                    + ldpaths["runpath"]
+                    + ldpaths["conf"]
+                    + ldpaths["interp"]
+                )
             realpath, fullpath = find_lib(elf, lib, all_ldpaths, root)
             _all_libs[lib] = {
-                'realpath': realpath,
-                'path': fullpath,
-                'needed': [],
+                "realpath": realpath,
+                "path": fullpath,
+                "needed": [],
             }
-            if fullpath:
-                lret = lddtree(realpath,
-                               root,
-                               prefix,
-                               ldpaths,
-                               display=fullpath,
-                               _first=False,
-                               _all_libs=_all_libs)
-                _all_libs[lib]['needed'] = lret['needed']
+            if realpath and fullpath:
+                lret = lddtree(
+                    realpath,
+                    root,
+                    prefix,
+                    ldpaths,
+                    display=fullpath,
+                    exclude=exclude,
+                    _first=False,
+                    _all_libs=_all_libs,
+                )
+                _all_libs[lib]["needed"] = lret["needed"]
 
         del elf
 
