@@ -5,15 +5,11 @@ import logging
 import os
 import pathlib
 import sys
-
-if sys.version_info[:2] >= (3, 8):
-    from importlib import metadata
-else:
-    import importlib_metadata as metadata
+from importlib import metadata
 
 import auditwheel
 
-from . import main_addtag, main_lddtree, main_repair, main_show
+from . import main_lddtree, main_repair, main_show
 
 
 def main() -> int | None:
@@ -40,7 +36,6 @@ def main() -> int | None:
     sub_parsers = p.add_subparsers(metavar="command", dest="cmd")
 
     main_show.configure_parser(sub_parsers)
-    main_addtag.configure_parser(sub_parsers)
     main_repair.configure_parser(sub_parsers)
     main_lddtree.configure_subparser(sub_parsers)
 
